#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kmem.h>
#include <sys/user.h>
#include <sys/proc.h>
#include <sys/cred.h>
#include <sys/disp.h>
#include <sys/buf.h>
#include <sys/vfs.h>
#include <sys/vnode.h>
#include <sys/fdio.h>
#include <sys/file.h>
#include <sys/uio.h>
#include <sys/conf.h>
#undef NFSCLIENT
#include <sys/statvfs.h>
#include <sys/mount.h>
#include <sys/pathname.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/sysmacros.h>
#include <sys/conf.h>
#include <sys/mkdev.h>
#include <sys/swap.h>
#include <sys/sunddi.h>
#include <sys/dktp/fdisk.h>
#include <fs/fs_subr.h>
#include <sys/modctl.h>
#include <sys/vol.h>
#include <sys/dkio.h>
#include <sys/open.h>
#include <sys/mntent.h>
#include <ext2fs.h>

#define MAXIMUM_PARTS	60
#define PART_TABLE_FLAG	0xaa55
#define table_check(b)	((unsigned short *)((b) + 0x1fe))
#define offset(b, n)	((struct partition *)((b) + 0x1be + \
				(n) * sizeof(struct partition)))
#define sector(s)	((s) & 0x3f)
#define cylinder(s, c)	((c) | (((s) & 0xc0) << 2))

#define calculate(h,s,c) (sector(s) - 1 + sectors * \
				((h) + heads * cylinder(s,c)))
#define set_hsc(h,s,c,sector) { \
				s = sector % sectors + 1;	\
				sector /= sectors;	\
				h = sector % heads;	\
				sector /= heads;	\
				c = sector & 0xff;	\
				s |= (sector >> 2) & 0xc0;	\
			}

#define ACTIVE_FLAG	0x80
#define EXTENDED	5

#define LINUX_PARTITION	0x81
#define LINUX_SWAP	0x82
#define LINUX_NATIVE	0x83

#define SECTOR_SIZE 512

	
/* from /usr/include/linux/genhd.h */
struct partition {
        unsigned char boot_ind;         /* 0x80 - active */
        unsigned char head;             /* starting head */
        unsigned char sector;           /* starting sector */
        unsigned char cyl;              /* starting cylinder */
        unsigned char sys_ind;          /* What partition type */
        unsigned char end_head;         /* end head */
        unsigned char end_sector;       /* end sector */
        unsigned char end_cyl;          /* end cylinder */
        unsigned int start_sect;        /* starting sector counting from 0 */
        unsigned int nr_sects;          /* nr of sectors in partition */
};

struct	partition *part_table[MAXIMUM_PARTS];

#define MAX_PER_LINE	16
void ext2fs_print_buffer(char buffer[], int n)
{
	int	i,
		l;
	char line[100];
	int  c=0;

        
	for (i = 0, l = 0; i < n; i++, l++) {
		if (l == 0)
		{
			sprintf(line,"0x%03X: ", i);
			c=7;
		}
		sprintf(&line[c]," %02X", (unsigned char) buffer[i]);
		c+=3;
		if (l == MAX_PER_LINE - 1) {
			line[c]='\0';
			printf("%s\n",line);
			l = -1;
		}
	}
	/*
	if (l > 0)
		printf("\n");*/
	printf("%s\n",line);
}

void ext2fs_print_part(struct partition *p)
{ 
  printf("AF  Hd Sec  Cyl  Hd Sec  Cyl   Start    Size ID\n");
  printf("%02x%4d%4d%5d%4d%4d%5d%8d%8d %02x\n",
				p->boot_ind, p->head,
				p->sector,
				p->cyl, p->end_head,
				p->end_sector,
				p->end_cyl,
				p->start_sect, p->nr_sects, p->sys_ind);
}

/*
 * ext2fs_whichpart()
 *  give a file spec of the form /dev/dsk/c1d0p2:3 returns the requested
 *  partition number and truncates the string to the real device name.
 *  So in this example returns 3 and /dev/dsk/c1d0p2. Returns 0 if no
 *  logical partition requested in string.
 */
int ext2fs_whichpart(char *spec)
{
  char *c;
  int  partno=0;
  
  /* is there a extended partition number on the end of the device name */
  c=strchr(spec,':');
  if (c!=NULL)
  {
    /* remember partition number - assumes partition number less than 10 ! */
    partno=(int)(*(c+1))-48;
    /* strip partition number from device name sting */
    *c='\0';  
  }
  
  return partno;
}

/*
 * ext2fs_get_start_sector():
 *  returns the start sector on the real device for the given partition number.
 *  This is done by reading the chain of extended partition tables, and counting
 *  the laogical native linux partitions found. Partitions are numbered from 1.
 *  returns 0 if the requested partition is not found or valid partition tables
 *  can't be found.
 */
int ext2fs_get_start_sector(struct ext2fs *fsp, int partno)
{
  struct buf *tp=NULL;
  int offset=0, sys_ind=0;
  char *buffer;
  int i;
  int part_start=0;
  int part_count=0;
  int found_part=0;
  
  do
  {
  
    tp = BREAD(fsp->e_devvp->v_rdev, offset,
	       SECTOR_SIZE);
    if (tp->b_flags & B_ERROR) 
    {
      goto failed;
    } 
    
    buffer=(char*)tp->b_un.b_addr;
  
    /* see of the buffer read is a partition table  */
    /* *** warning *** this test is not endian safe */
    if (*(unsigned short *)  (0x1fe + buffer) != 0xAA55) 
    {
      cmn_err(CE_WARN,"ext2fs_get_start_sector: offset %d, not a valid partition table\n",offset); 
      goto failed;	/*  not a valid partition table */  
    }
  		
    for (i=0;i<4;i++)
    {
      part_table[i]=offset(buffer,i);
      /* ext2fs_print_part(part_table[i]); */

      if (part_table[i]->sys_ind == LINUX_NATIVE)
      {
        part_count++;
	if (part_count==partno)
	{
	  /* work out the start sector of the requested partition */
	  part_start = part_table[i]->start_sect + offset;
	  found_part=1;
	}
      }
    }
    
    /* move to next partition table in chain */
    offset=part_table[0]->start_sect;
    sys_ind=part_table[0]->sys_ind;
    
    /* free the block */
    tp->b_flags |= B_STALE | B_AGE;
    brelse(tp);
    
  } while ((sys_ind == EXTENDED) && (found_part==0));	   
           
  /* printf("part %d starts at %d\n",partno,part_start); */
  
  return part_start;
  
failed:
  if (tp) 
  {
    tp->b_flags |= B_STALE | B_AGE;
    brelse(tp);
  }
  
  return 0;
}
 
